#include "types.t"
#include "defines.t"

// CLASS GENERIC

class generic
{
  area "sys";

  inherits generic_view;

#include "check_acl.t"

  container_class '';
  container_column_name '';

  include_c "http_protocol.h";
  include_c "http_main.h";
  include_c "co_session.h";

  default_form "showForm";

 // PROPERTIES

#include "generic_props.t"

// METHODS

//TODO: What if not found?

  method get
  {
   id_permission __PERMISSION_GET;
   sec_type __METHOD_SEC_TYPE_OBJECT;

   doc "this method retrieves data for given object. Object is identified using par_area, par_class and par_id"
   {
       pl "Ta metoda zwraca dane dla obiektu identyfikowanego przez par_area, par_class i par_id";
   };
   
    __PARAMS_HEADER_WITH_CALLBACK
    
    body language_c
      {
         C
	 %%
         __CREATE_QUERY("SELECT * FROM cov_%s_%s WHERE id=%d" , par_area, par_class,par_id)
         __GET_RESULT_AND_CREATE_JS
	 %%
	 ;
      };  
								    
  };

  method getProp
  {
   id_permission __PERMISSION_GET;
   sec_type __METHOD_SEC_TYPE_OBJECT;

   doc "This method retrieves value (CO_STRING) of given property for given object. Object is identified using par_area, par_class and par_id; property using prop_name."
   {
       pl "Ta metoda zwraca warto (jako CO_STRING) property dla obiektu identyfikowanego przez par_area, par_class i par_id; nazw property podajemy w prop_name.";
   };
   
    __PARAMS_HEADER_WITH_CALLBACK

    param prop_name 
    {
      type CO_STRING;
      doc "Name of prop to retrieve"
      {
         pl "Nazwa property, ktrego warto mamy otrzyma";
      };
    };

    declare language_c
    {
      C
      %%
         const char* prop_value = "";
      %%
      ;
    };
    
    body language_c
      {
         C
	 %%
    	    __SQL_GET_STRING(prop_value, "SELECT \"%s\" FROM cov_%s_%s WHERE id=%d" , prop_name, par_area, par_class,par_id)
	    __CREATE_JS_FROM_STRING(prop_value)
	    __RETURN_STRING(prop_value)
	 %%
	 ;
      };  
								    
  };

  method getChildrenCount
  {
   id_permission __PERMISSION_GET_RELATED;
   sec_type __METHOD_SEC_TYPE_OBJECT;

   doc "This method retrieves number (CO_INTEGER) of (non-deleted) children given property for given object. Object is identified using child_class, child_prop and foreign_prop."
   {
       pl "";
   };
   
    __PARAMS_HEADER_WITH_CALLBACK

    param child_class 
         { 
           type CO_STRING; 
           doc "this is the name of the class of children counted";
         };
    param child_prop 
        { 
           type CO_STRING;
           doc "this is the property name in child class that holds id of parent";
        };
    param foreign_prop 
       { 
           type CO_STRING;
           doc "this is the prop in container (this:) class that holds id to copy to child class";
       };

    declare language_c
    {
      C
      %%
         int children_count = 0;
      %%
      ;
    };
    
    body language_c
      {
         C
	 %%
		__SQL_GET_INTEGER(children_count, 
			"SELECT count(*) FROM co_%s_%s AS C, co_%s_%s AS P WHERE P.id=%d and P.deleted=0 and C.\"%s\"=P.\"%s\"" , 
				par_area, child_class,
				par_area, par_class,
				par_id,
				child_prop, foreign_prop)
	    __CREATE_JS_FROM_INTEGER(children_count)
	    __RETURN_INTEGER(children_count)
	 %%
	 ;
      };  
								    
  };

  method getCatalogWrapper
  {
   id_permission __PERMISSION_GET_RELATED;
   sec_type __METHOD_SEC_TYPE_OBJECT;

   doc "This method retrieves id (CO_INDEX) of the catalog wrapper. Requires catalog class name. Returns -1 if not found."
   {
       pl "";
   };
   
    __PARAMS_HEADER_WITH_CALLBACK

    param catalog_class 
         { 
           type CO_STRING; 
           doc "this is the name of the catalog class";
         };

    declare language_c
    {
      C
      %%
         int id_catalog_wrapper = -1;
      %%
      ;
    };
    
    body language_c
      {
         C
	 %%
		__SQL_GET_INTEGER(id_catalog_wrapper, 
			"SELECT C.id FROM co_%s_%s AS C, co_%s_catalog_type AS CT "
			"WHERE C.id_item=%d and C.deleted=0 and C.id_catalog_type=CT.id "
			" and CT.deleted=0 and CT.parent_class='%s' and CT.class_name='%s'", 
				par_area, catalog_class, par_area,
				par_id,
				catalog_class, par_class)
	    __CREATE_JS_FROM_INTEGER(id_catalog_wrapper)
	    __RETURN_INTEGER(id_catalog_wrapper)
	 %%
	 ;
      };  
								    
  };

//TODO: What if not found?

 method increaseVersion
 {
    id_permission __PERMISSION_UPDATE;
    sec_type __METHOD_SEC_TYPE_OBJECT;

    doc "Increases version number and sends action with the new version number. Checks whether object is deleted (updates only not deleted; returns -1 otherwise)."
    {
        pl "Zwieksza numer wersji dla danego obiektu i wysyla akcje CO_OP_UPDATE, version = version+1. Sprawdza czy obiekt nie jest w stanie deleted, jesli jest to zwraca -1.";
    };

    __PARAMS_HEADER
    
    declare language_c
    {
      C "int cur_version = 0;";
      C "const char* puta_classes[2]={NULL,NULL};";
      C "int puta_ids[2]={0,0}; // for queueActionX";
    };

    body language_c
    {
      C
      %%

      __CREATE_QUERY("UPDATE co_%s_%s SET version=version+1 WHERE id=%d AND deleted=0",
                      par_area, par_class, par_id);
      __SQL_EXEC
      __ASSERT(ok, "Object is deleted or there is no such object")

      __SQL_GET_INTEGER(cur_version, "SELECT version FROM co_%s_%s WHERE id=%d", par_area, par_class, par_id)
      __ASSERT(ok, "Cannot get version number")

      puta_classes[0] = par_class;
      puta_ids[0] = par_id;

      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, "version",ap_psprintf(p,"%d",cur_version),NULL);
      %%
      ;
    };
   };
 
 method update
 {
   id_permission __PERMISSION_UPDATE;
   sec_type __METHOD_SEC_TYPE_OBJECT;

   doc "This method updates one prop of one object. Sends action CO_OP_UPDATE"
   {
     pl "Ta metoda zmienia jeden prop danego obiektu. Wysyla akcje CO_OP_UPDATE";
   };

    __PARAMS_HEADER_WITH_VERSION
    param par_prop 
    { 
      type CO_STRING; 
      doc "Name of the prop to be updated"
      {
         pl "Nazwa prop, ktoremu bedzie zmieniona wartosc";
      };
    };
    param new_value 
    { 
      type CO_STRING; 
      doc "New value for prop. All values are passed as CO_STRING type!"
      {
         pl "Nowa wartosc. Wszystkie wartosci sa przekazywane jako typ CO_STRING !";
      };
    };
    
    declare language_c
    {
      C "int cur_version = 0;";
      C "const char *new_value_coded;";
      C "const char* puta_classes[2]={NULL,NULL};";
      C "int puta_ids[2]={0,0}; // for queueActionX";
    };

    body language_c
    {
      C
      %%
      __CHECK_OBJECT_VERSION("Nie poprawilismy danych bo miales stara wersje obiektu. Odswiez swojego klienta i sprobuj jeszcze raz") 

      new_value_coded = cou_buildEncodedSQL(cnx, new_value);
      __CREATE_QUERY("UPDATE co_%s_%s SET \"%s\"=\'%s\', version=version+1 WHERE id=%d ",
                      par_area, par_class,par_prop, new_value_coded,par_id);
      __SQL_EXEC
      __ASSERT(ok, "Error during updating data, probably there is no such object or prop: par_area '%s' ,par_class '%s', par_id %d, par_prop '%s', new_value '%s'", 
                    par_area, par_class, par_id, par_prop, new_value)
     
      puta_classes[0] = par_class;
      puta_ids[0] = par_id;
      
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, par_prop, new_value, NULL);
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, "version",ap_psprintf(p,"%d",cur_version+1),NULL);
      %%
      ;
    };
   };
 
 method multiUpdate
 {
   id_permission __PERMISSION_UPDATE;
   sec_type __METHOD_SEC_TYPE_OBJECT;

   doc "This method updates properties of one object. Sends actions CO_OP_UPDATE"
   {
     pl "Ta metoda zmienia atrybut(y) danego obiektu. Wysyla akcje CO_OP_UPDATE";
   };

    __PARAMS_HEADER_WITH_VERSION
    param par_props
    { 
      type CO_ARRAY_OF_STRINGS;
      doc "Name of the properties to be updated"
      {
         pl "Nazwy atrybutw, ktorym bdzie zmieniona warto";
      };
    };
    param new_values
    { 
      type CO_ARRAY_OF_STRINGS;
      doc "New values for properties. All values are passed as CO_STRING type!"
      {
         pl "Nowe wartoci. Wszystkie wartoci sa przekazywane jako typ CO_STRING !";
      };
    };
    
    declare language_c
    {
      C "int i;";
      C "int cur_version = 0;";
      C "const char *new_value_coded;";
      C "const char* puta_classes[2]={NULL,NULL};";
      C "int puta_ids[2]={0,0}; // for queueActionX";
      C "int NProps;";
      C "int NValues;";
      C
      %%
      const char *strSet = "";
      %%
      ;
    };

    body language_c
    {
      C
      %%

      __CHECK_OBJECT_VERSION("Nie poprawilimy danych bo miale star wersj obiektu. Odswie swojego klienta i sprobuj jeszcze raz")

      __SEND_MESG_DEBUG("1")


      NProps = coa_GetElementCount(cnx, par_props);
      
      __SEND_MESG_DEBUG("NProps=%d",NProps)
      
      NValues = coa_GetElementCount(cnx, new_values);

      __SEND_MESG_DEBUG("NValues=%d",NValues)


      __ASSERT(NProps=NValues,"the number of properties doesn't match the number of values")
      __ASSERT(NProps>0,"no properties, no values")

      __SEND_MESG_DEBUG("4")


      for (i=0;i<NProps;i++)
      {
        new_value_coded = cou_buildEncodedSQL(cnx, coa_GetElement(cnx, new_values, i));
        strSet = ap_pstrcat(p, strSet, "\"", coa_GetElement(cnx, par_props, i), "\"=\'", new_value_coded, "\', ",NULL);
      }

      __SEND_MESG_DEBUG("5")


      __CREATE_QUERY("UPDATE co_%s_%s SET %sversion=version+%d WHERE id=%d ",
                      par_area, par_class, strSet, NProps, par_id);
      __SQL_EXEC
      __ASSERT(ok, "Error during updating data, probably there is no such object or prop: par_area '%s' ,par_class '%s', par_id %d", 
                    par_area, par_class, par_id)
      
      puta_classes[0] = par_class;
      puta_ids[0] = par_id;
      
      for (i=0; i<NProps; i++) cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, coa_GetElement(cnx, par_props, i), coa_GetElement(cnx, new_values, i), NULL);
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, "version",ap_psprintf(p,"%d",cur_version+NProps),NULL);
      %%
      ;
    };
   };


 method delete
 {
   id_permission __PERMISSION_DELETE;
   sec_type __METHOD_SEC_TYPE_OBJECT;
   
   doc "Sets deleted prop to 1. Sends action CO_OP_DELETE."
   {
      pl "Ustawia prop deleted na 1. Wysyla akcje CO_OP_DELETE.";
   };

    __PARAMS_HEADER_WITH_VERSION
   
    declare language_c
    {
      C
      %%
      int cur_version = 0;
      const char *new_value_coded;
      const char* puta_classes[2]={NULL,NULL};
      int puta_ids[2]={0,0}; // for queueActionX
      %%
      ;
    };

    body language_c
    {
      C 
      %%
      __CHECK_OBJECT_VERSION("Nie skasowalismy obiektu bo miales stara wersje obiektu. Odswiez swojego klienta i sprobuj jeszcze raz")

      __CREATE_QUERY("UPDATE co_%s_%s SET deleted=1, version=version+1 WHERE id=%d ",
              par_area, par_class,par_id);
      __SQL_EXEC
      __ASSERT(ok, "Error during deleting object, probably there is no such object: par_area '%s' ,par_class '%s', par_id %d ", 
                    par_area, par_class, par_id)
    

      puta_classes[0] = par_class;
      puta_ids[0] = par_id;
      
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_DELETE, "deleted", ap_psprintf(p,"%d",1), NULL);
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, "version",ap_psprintf(p,"%d",cur_version+1),NULL);
      %%
      ;
    };
   };

 method undelete
 {
   id_permission __PERMISSION_UNDELETE;
   sec_type __METHOD_SEC_TYPE_OBJECT;
 
   doc "Sets deleted prop to 0. Sends action CO_OP_UNDELETE."
   {
      pl "Ustawia prop deleted na 0. Wysyla akcje CO_OP_UNDELETE.";
   };

  
    __PARAMS_HEADER_WITH_VERSION
    
    declare language_c
    {
      C "int cur_version = 0;";
      C "const char* puta_classes[2]={NULL,NULL};";
      C "int puta_ids[2]={0,0}; // for queueActionX";

    };  

    body language_c
    {
      C 
      %%
      __CHECK_OBJECT_VERSION("Nie odkasowalimy obiektu bo miales stary obraz! (bez obrazy :-)")
      __CREATE_QUERY("UPDATE co_%s_%s SET deleted=0, version=version+1 WHERE id=%d ",
              par_area, par_class,par_id);
      __SQL_EXEC
      __ASSERT(ok, "Error during undeleting object, probably there is no such object: par_area '%s' ,par_class '%s', par_id %d ", 
                    par_area, par_class, par_id)
 
      puta_classes[0] = par_class;
      puta_ids[0] = par_id;
      
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UNDELETE, "deleted", ap_psprintf(p,"%d",0), NULL);
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_UPDATE, "version",ap_psprintf(p,"%d",cur_version+1),NULL);
      %%
      ;
    };
   };
 
 method insertChild
 {
    id_permission __PERMISSION_INSERT_CHILD;
    sec_type __METHOD_SEC_TYPE_OBJECT;
   
    doc "this is method for inserting object into container. You can specify class_name of child element";

    __PARAMS_HEADER    
    param child_class 
         { 
           type CO_STRING; 
           doc "this is the name of class tu insert record in";
         };
    param child_prop 
        { 
           type CO_STRING;
           doc "this is the property name in child class that holds id of parent";
        };
    param foreign_prop 
       { 
           type CO_STRING;
           doc "this is the prop in container (this:) class that holds id to copy to child class";
       };
    
    declare language_c
    {
      C 
      %%
        int new_id=0;
	int i;
	int nFields;
        const char* puta_classes[3]={NULL,NULL,NULL};
	int puta_ids[3]={0,0,0}; // for queueActionX 
	const char **props;
      %%
      ;	
    };  
    
    body language_c
    {
      C 
      %%
      __CREATE_QUERY("INSERT INTO co_%s_%s(%s) SELECT %s FROM co_%s_%s WHERE id = %d",
                           par_area, child_class, child_prop, foreign_prop, par_area, par_class, par_id)

      __SQL_EXEC
      __ASSERT(ok, "Error during inserting object: par_area '%s' ,par_class '%s', par_id %d, child_class '%s', child_prop '%s', foreign_prop '%s' ", 
                    par_area, par_class, par_id, child_class, child_prop, foreign_prop)
 
      __SQL_GET_INTEGER(new_id,"SELECT last_value FROM co_%s_%s_sqid", 
                           par_area, child_class)

      __SEND_MESG_DEBUG("generic:insertChild done id=%d",new_id)
      
      puta_classes[0] = par_class;
      puta_ids[0] = par_id;

      puta_classes[1] = child_class;
      puta_ids[1] = new_id;
      
      cou_queueActionX( cnx, par_class, puta_classes, puta_ids, CO_OP_INSERT, "child_class", child_class,NULL);

      // getting current values

      __CREATE_QUERY("SELECT * FROM cov_%s_%s WHERE id = %d",
                           par_area, child_class, new_id)
      __SQL_GET_RESULT
      __ASSERT(ok, "Could not get data for new object (just inserted), par_area '%s', child_class '%s', new_id %d",
                     par_area, child_class, new_id) 
      
      puta_classes[0] = child_class;
      puta_ids[0] = new_id;

      puta_classes[1] = par_class;
      puta_ids[1] = par_id;

      puta_classes[2] = NULL;
      puta_ids[2] = 0;
      
      nFields = cor_GetFieldCount(cnx, corec);
      props = ap_palloc(p, 2 * nFields * sizeof(char*) + 1);
      
      for (i = 0; i < nFields; i++)
      {
	 props[2*i] = ap_pstrdup(p,   cor_GetFieldName(cnx, corec, i));
	 props[2*i+1] = ap_pstrdup(p, cor_GetValue    (cnx, corec, 0, i));
      } 
      props[2*i]=NULL;
      
      //cou_sendMessage(cnx, "we have props");
      
      __CLEAR_RESULT
      cou_queueActionXX( cnx, child_class, puta_classes, puta_ids, CO_OP_UPDATE, props);  
      cou_setReturnValue(cnx, &new_id, CO_INTEGER);
      
      %%
      ;
    };
   };


	method getRelatedData
	{

		doc "retrieves data of related class";

		id_permission __PERMISSION_GET_RELATED;
		sec_type __METHOD_SEC_TYPE_OBJECT;
     
		__PARAMS_HEADER_WITH_CALLBACK
		param related_class 
		{
			type CO_STRING;
			doc "Name of the class to retrieve related data from. Current class (par_class) should be container for related_class"
			{
				pl "Nazwa klasy, dla ktrej pobieramy dane. Bieca klasa (par_class) powinna by pojemnikiem dla related_class";
			};
		}; 

		param related_prop
		{
			type CO_STRING;
			doc "Name of prop in the related class used to form relation"
			{
				pl "Nazwa prop w klasie pojemnika, ktre jest uywane do wizania relacji (najczciej id)";
			};
			optional 1;
			default "related_prop=NULL";
		}; 

		param dynamic_prop
		{
			type CO_STRING;
			doc "If not null is name of prop where is stored dynamic name of container class"
			{
				pl "Jeli nie jest null, to jest to nazwa prop, w ktrej jest przechowywana dynamieczna nazwa klasy pojemnika";
			};
			optional 1;
			default "dynamic_prop=NULL";
		}; 


		declare language_c
		{
			C
			%%
			const char* foreign_prop;
			char* foreign_val = NULL;
			const char * rel_prop;
			%%
			;
		};
       
		body language_c
		{
			C 
			%%
  
			// getting foreign_prop
			if (related_prop)
			{
				__CREATE_QUERY("SELECT prop_foreign_prop FROM co_%s_meta_prop " 
				               " WHERE (prop_foreign='%s' OR prop_foreign_dynamic_classes LIKE '%%|%s|%%') AND class_name='%s' AND prop_name='%s'",
				                par_area, par_class, par_class, related_class, related_prop)
				__SQL_GET_RESULT
				// ok == TupleCount
				__ASSERT(ok>0, "Could not find related prop for area:'%s' class:'%s' related_class:'%s'",
				                par_area,par_class, related_class)
				rel_prop=related_prop;
			}
			else
			{
				__CREATE_QUERY("SELECT prop_foreign_prop, prop_name FROM co_%s_meta_prop "
				               " WHERE (prop_foreign='%s' OR prop_foreign_dynamic_classes LIKE '%%|%s|%%') and class_name='%s'",
				               par_area, par_class, par_class, related_class )
				__SQL_GET_RESULT
				// ok == TupleCount
				__ASSERT(ok>0, "Could not find related prop for area:'%s' class:'%s' related_class:'%s'",
				                par_area,par_class, related_class)
				rel_prop = cor_GetCopyOfValueByName(cnx, corec, 0, "prop_name"); 
			}

			foreign_prop = cor_GetCopyOfValueByName(cnx, corec, 0, "prop_foreign_prop"); 
			__CLEAR_RESULT
	  
			// getting foreign_val
			__SQL_GET_STRING(foreign_val, "SELECT %s FROM cov_%s_%s WHERE id=%d", foreign_prop, par_area, par_class, par_id)
			__ASSERT(ok, "Could not get foreign_val, SQL: SELECT %s FROM cov_%s_%s WHERE id=%d", foreign_prop, par_area, par_class, par_id)
		    
			// getting data
			if(dynamic_prop)
			{
				__CREATE_QUERY("SELECT * FROM cov_%s_%s WHERE %s='%s' AND %s='%s'", par_area, related_class, rel_prop, foreign_val, dynamic_prop, par_class)
			}
			else
			{
				__CREATE_QUERY("SELECT * FROM cov_%s_%s WHERE %s='%s' ORDER BY id", par_area, related_class, rel_prop, foreign_val)
			}
			__GET_RESULT_AND_CREATE_JS
			
			%%
			;
		}; // body
	}; // method


 method getWithSchema
   {
      doc  "Retrieves schema for given class";

    id_permission __PERMISSION_GET;
    sec_type __METHOD_SEC_TYPE_OBJECT;
    
     __PARAMS_HEADER
     param callbacks { type CO_ARRAY_OF_STRINGS; };

      body language_c
       {
          C 
	  %%
//
	  @co_call(par_area,par_class,"getSchema",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,0)', representant_level=>-1);
	  @co_call(par_area,par_class,"get",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx, callbacks, 1)');
	  %%
	  ;
       };
   };

 method getSchemaRelatedAndData
   {
      doc  "Retrieves schema related tables and data for given class element";

    id_permission __PERMISSION_GET_RELATED;
    sec_type __METHOD_SEC_TYPE_OBJECT;
    
     __PARAMS_HEADER
     param callbacks { type CO_ARRAY_OF_STRINGS; };

      body language_c
       {
          C 
	  %%
	  
	  // HERE WE SHOULD CHECK THE LENGTH OF callback
	  
	  @co_call(par_area,par_class,"getSchema",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,0)', \
                   representant_level=>-1);	  
	  @co_call(par_area,par_class,"getRelatedClasses",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,1)');	  
	  @co_call(par_area,par_class,"get",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,2)');
	  %%
	  ;
       };
   };

 method getWithSchemaX
   {
    doc  "Retrieves schema for given class";

    id_permission __PERMISSION_GET;
    sec_type __METHOD_SEC_TYPE_OBJECT;

    __PARAMS_HEADER
    param callbacks { type CO_ARRAY_OF_STRINGS; };

    body language_c
      {
	  C
	  %%
//
	  @co_call(par_area,par_class,"getSchemaX",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,0)', deep_level=>2);
	  @co_call(par_area,par_class,"get",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,1)');
	  %%
	  ;
      };
   };

 method getSchemaXRelatedAndData
   {
      doc  "Retrieves schema related tables and data for given class element";

    id_permission __PERMISSION_GET_RELATED;
    sec_type __METHOD_SEC_TYPE_OBJECT;
    
     __PARAMS_HEADER
     param callbacks { type CO_ARRAY_OF_STRINGS; };
     param deep_level{ type CO_INTEGER; optional 1; default "deep_level=3;"; };
 
     declare language_c
     {
        C 'co_array_of_pointers * get_class_names    = coa_Create(cnx);';
        C 'co_array_of_pointers * get_prop_names     = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_class_names = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_prop_names  = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_operators   = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_values      = coa_Create(cnx);';
        C 'int NProps = 0;';
        C 'int i = 0;';
     };
 
      body language_c
       {
          C 
	  %%
	  if(deep_level<0) deep_level=0;

          coa_AppendElement(cnx, filter_class_names, ap_psprintf(p, "%s",par_class));
          coa_AppendElement(cnx, filter_prop_names,  ap_psprintf(p, "id"));
          coa_AppendElement(cnx, filter_operators,   ap_psprintf(p, "="));
          coa_AppendElement(cnx, filter_values,      ap_psprintf(p, "%d", par_id));
  
	  // HERE WE SHOULD CHECK THE LENGTH OF callback
	  
          @co_call(par_area,par_class,"getSchemaX",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,0)', deep_level=>deep_level); 

          // READING VALUE FOR CURRENT OBJECT
          corec = cou_getReturnValue(cnx, CO_RECORDSET, &err);
          __ASSERT(err==0, "Error getting recordset value")

          NProps = cor_GetTupleCount(cnx, corec);

          for(i=0; i<NProps; i++)
          {
             coa_AppendElement(cnx, get_class_names, ap_psprintf(p, par_class));
             coa_AppendElement(cnx, get_prop_names,  cor_GetCopyOfValueByName(cnx, corec, i, "get_prop_name"));
          }
          @co_call(par_area,par_class,"selectX",par_area=>par_area,par_class=>par_class,par_id=>-1,callback=>'coa_GetElement(cnx,callbacks, 2)', \
                  get_class_names=>get_class_names,get_prop_names=>get_prop_names, \
                  filter_class_names=>filter_class_names,filter_prop_names=>filter_prop_names, \
                  filter_operators=>filter_operators, filter_values=>filter_values);

          //GETTING RELATED CLASSES
 	  @co_call(par_area,par_class,"getRelatedClasses",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,1)');	  
	  %%
	  ;
       };
   };


	method getForRelatedSchemaAndData_old
	{
		doc  "Retrieves schema related tables and data for given class element";

		id_permission __PERMISSION_GET_RELATED;
		sec_type __METHOD_SEC_TYPE_CLASS;

    
		__PARAMS_HEADER
		param related_class {type CO_STRING;}; 
		param related_prop {type CO_STRING;}; 
		param dynamic_prop
		{
			type CO_STRING;
			doc "If not null is name of prop where is stored dynamic name of container class"
			{
				pl "Jeli nie jest null, to jest to nazwa prop, w ktrej jest przechowywana dynamiczna nazwa klasy pojemnika";
			};
			optional 1;
			default "dynamic_prop=NULL";
		}; 
		param callbacks { type CO_ARRAY_OF_STRINGS; };

		body language_c
		{
			C 
			%%
	  
	  // HERE WE SHOULD CHECK THE LENGTH OF callbacks
	  
     @co_call(par_area,related_class,"getSchema", par_area=>par_area,par_class=>related_class,par_id=>-1,callback=>"coa_GetElement(cnx,callbacks,0)", representant_level=>-1);	  
     @co_call(par_area,par_class,"getRelatedData",par_area=>par_area,par_class=>par_class,par_id=>par_id,related_class=>related_class,related_prop=>related_prop,callback=>"coa_GetElement(cnx,callbacks, 1)", dynamic_prop=>dynamic_prop);
			%%
			;
		};
	};

 method getForRelatedSchemaAndData
   {
      doc  "Retrieves schema related tables and data for given class element";

    id_permission __PERMISSION_GET_RELATED;
    sec_type __METHOD_SEC_TYPE_OBJECT;
    
     __PARAMS_HEADER
		param related_class {type CO_STRING;}; 
		param related_prop {type CO_STRING;}; 
		param dynamic_prop
		{
			type CO_STRING;
			doc "If not null is name of prop where is stored dynamic name of container class"
			{
				pl "Jeli nie jest null, to jest to nazwa prop, w ktrej jest przechowywana dynamiczna nazwa klasy pojemnika";
			};
			optional 1;
			default "dynamic_prop=NULL";
		}; 
     param callbacks { type CO_ARRAY_OF_STRINGS; };
     param deep_level{ type CO_INTEGER; optional 1; default "deep_level=0;"; };
 
     declare language_c
     {
        C 'co_array_of_pointers * get_class_names    = coa_Create(cnx);';
        C 'co_array_of_pointers * get_prop_names     = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_class_names = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_prop_names  = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_operators   = coa_Create(cnx);';
        C 'co_array_of_pointers * filter_values      = coa_Create(cnx);';
        C 'int NProps = 0;';
        C 'int i = 0;';
        C 'const char * prop_name = "";';
        C 'const char * prop_value = "";';
     };
 
      body language_c
       {
          C
	  %%
	  if(deep_level<0) deep_level=0;

/**/
		 /* get foreign prop for related prop */
		 /* should be in meta_prop as method: */
         __SQL_GET_STRING( prop_name, "
SELECT P.prop_foreign_prop FROM co_sys_meta_prop P
WHERE P.area='%s' AND P.class_name='%s' AND P.prop_name='%s' AND P.deleted=0" ,
                    par_area, related_class, related_prop)
          @co_call(par_area,par_class,"getProp",par_area=>par_area,par_class=>par_class,par_id=>par_id,prop_name=>prop_name);
		  __GET_RETURNED_STRING(prop_value)
/**/
          coa_AppendElement(cnx, filter_class_names, ap_psprintf(p, "%s", related_class));
          coa_AppendElement(cnx, filter_prop_names,  ap_psprintf(p, "%s", related_prop));
          coa_AppendElement(cnx, filter_operators,   ap_psprintf(p, "="));
          coa_AppendElement(cnx, filter_values,      ap_psprintf(p, "%s", prop_value));

	  if (dynamic_prop)
	  {
            coa_AppendElement(cnx, filter_class_names, ap_psprintf(p, "%s", related_class));
            coa_AppendElement(cnx, filter_prop_names,  ap_psprintf(p, "%s", dynamic_prop));
            coa_AppendElement(cnx, filter_operators,   ap_psprintf(p, "="));
            coa_AppendElement(cnx, filter_values,      ap_psprintf(p, "%s", par_class));
	  }

	  // HERE WE SHOULD CHECK THE LENGTH OF callback

          @co_call(par_area,par_class,"getSchemaX",par_area=>par_area,par_class=>related_class,par_id=>-1,callback=>'coa_GetElement(cnx,callbacks,0)', deep_level=>deep_level);

          // READING VALUE FOR CURRENT OBJECT
		  __GET_RETURNED_RECORDSET(corec)
          //corec = cou_getReturnValue(cnx, CO_RECORDSET, &err);
          __ASSERT(err==0, "Error getting recordset value")

          NProps = cor_GetTupleCount(cnx, corec);

          for(i=0; i<NProps; i++)
          {
             coa_AppendElement(cnx, get_class_names, ap_psprintf(p, related_class));
             coa_AppendElement(cnx, get_prop_names,  cor_GetCopyOfValueByName(cnx, corec, i, "get_prop_name"));
          }
          @co_call(par_area,related_class,"selectX",par_area=>par_area,par_class=>related_class,par_id=>-1,callback=>'coa_GetElement(cnx,callbacks, 1)', \
                  get_class_names=>get_class_names,get_prop_names=>get_prop_names, \
                  filter_class_names=>filter_class_names,filter_prop_names=>filter_prop_names, \
                  filter_operators=>filter_operators, filter_values=>filter_values);

          //GETTING RELATED CLASSES
 	  //co_call(par_area,par_class,"getRelatedClasses",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>'coa_GetElement(cnx,callbacks,1)');	  
	  %%
	  ;
       };
   };

 method showForm
   {

     id_permission __PERMISSION_GET;
     sec_type __METHOD_SEC_TYPE_OBJECT;
     
     __PARAMS_HEADER
     param deep_level
     {
       optional 1;
       default "deep_level=0";
       type CO_INTEGER;
     };

		param context_name
		{
			doc "Name of context in which we execute this form. Now this is the name of additionall menu";
			type CO_STRING;
			optional 1;
			default "context_name=ap_pstrdup(p, \"\");";
		};

		param context1
		{
			doc "Additionall parameter passed to menu";
			type CO_STRING;
			optional 1;
			default "context1=ap_pstrdup(p, \"\");";
		};

		param context2
		{
			doc "Additionall parameter passed to menu";
			type CO_STRING;
			optional 1;
			default "context2=ap_pstrdup(p, \"\");";
		};

		param context3
		{
			doc "Additionall parameter passed to menu";
			type CO_STRING;
			optional 1;
			default "context3=ap_pstrdup(p, \"\");";
		};

     declare language_c
      {
        C
	%%
	FILE *fh;
	int c;
        const char* root_path;
        const char* class_label="";
        long len;
        int user_id=1;
	char * name_role=NULL;
	char* block_date = "1900-01-01";
	%%
	;
      };

     body language_c
      {
        C
	%%
        root_path = cou_getCORootPath( cnx );

		__SQL_GET_STRING(name_role, "SELECT name FROM co_sys_role WHERE id=%d", cnx->id_role)
		if(strcmp(par_area,"sys")!=0) {
			
		  __SQL_GET_STRING(block_date, "SELECT block_date FROM co_%s_block_date WHERE id=1", par_area)
			
		}


        fh = ap_pfopen(p, ap_pstrcat( p, root_path, "/forms/", "generic", "_head.html", NULL), "r");
        if (!fh) cou_error( cnx, "Head file not found");
        cou_logError( cnx, CO_LOG_DEBUG, "sending head" );
        ap_rputs("<HEAD>\n", r);
        ap_kill_timeout(r);

        len = ap_send_fd( fh, r ); /* returns total of bytes sent */
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Head sent: %ld bytes", len);

        ap_pfclose(p,fh);


        cou_logError( cnx, CO_LOG_DEBUG, "sending script" );
        ap_rputs("<SCRIPT>\n", r);

	ap_rprintf( r, "var  id_session = %d;\n", cou_assignNewSession( cnx ) );
	ap_rprintf( r, "var  par_id = %d;\n", par_id );
	ap_rprintf( r, "var  par_class=\'%s\';\n", par_class );
	ap_rprintf( r, "var  par_area = \'%s\';\n", par_area );
	ap_rprintf( r, "var  debug_level = %d;\n", cnx->debug_level );
	ap_rprintf( r, "var  deep_level = %d;\n", deep_level );
	ap_rprintf( r, "var  language = \'%s\';\n", cnx->language );
	ap_rprintf( r, "var  role = {id : %d , name: \'%s\'};\n", cnx->id_role, name_role );
	ap_rprintf( r, "var  block_date = String(\'%s\');\n", block_date );
			ap_rprintf( r, "var  context_name = \'%s\';\n", context_name );
			ap_rprintf( r, "var  context1 = \'%s\';\n", context1 );
			ap_rprintf( r, "var  context2 = \'%s\';\n", context2 );
			ap_rprintf( r, "var  context3 = \'%s\';\n", context3 );

	__SQL_GET_STRING(class_label, "
SELECT L.text FROM co_sys_meta_class AS M, co_sys_label_class AS L 
WHERE M.area = '%s' AND M.class_name = '%s' AND L.id_meta_class = M.id AND L.language = '%s'
",		par_area,
		par_class, cnx->language)

	ap_rprintf( r, "var  class_label = \'%s\';\n", class_label );

        ap_rputs("</SCRIPT>\n", r);

        ap_rputs("</HEAD>\n", r);
        fh = ap_pfopen(p, ap_pstrcat( p, root_path, "/forms/", "generic", "_body.html", NULL), "r");
        if (!fh) cou_error( cnx, "Body file not found");
        cou_logError( cnx, CO_LOG_DEBUG, "sending body" );
        ap_rputs("<BODY>\n", r);
        len = ap_send_fd( fh, r ); /* returns total of bytes sent */

        if (r->connection->aborted) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "connection aborted");
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "bytes_send=%ld", r->bytes_sent);

        ap_soft_timeout((char *) "finish request", r);

        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Body sent: %ld bytes", len);

        ap_pfclose(p,fh);
        ap_rputs("</BODY>\n", r);

        ap_rflush(r);
        %%
        ;
    };
  };
  
};

//END OF CLASS GENERIC
