#include "types.t"      
#include "defines.t"      

// CLASS CATALOG

class js_code

{ 
  area "sys";
  
  inherits generic_acl;

#include "check_acl.t"
  
  container_class '';
  container_column_name '';
      
 // PROPERTIES
      
#include "generic_acl_props.t"      

  prop code    
  {      
    type CO_STRING;      
    default "''";      
    params 'ROWS=20;COLS=60';      
    representant 100;      
    foreign '';      
    foreign_prop '';      
    position 200;    
  };    

	prop name
	{
  f_body
  %%
     SELECT ''js_code['' || id($1) || '']'';
  %%
  ;
		type CO_STRING;
		default "''";
		params 'READONLY=1';
		representant 0;
		foreign '';
		foreign_prop '';
		position 400;
		group_path '';
   };
 

// METHODS
  uses "js_lib";
  uses "catalog_type";

  include_c "jsapi.h";
  include_c "jsatom.h";
  include_c "co_mod.h";
  include_c "../../coffice/co_js.h";
  include_c "../../coffice/ap_parser.h";

  local_method loc_lineTranslation
  {
    doc
    %%
	Translate line numbers in asp-generated prolog.
	For all other code return original line number.
    %%
    ;

    C
    %%
     int loc_lineTranslation
            (
		const char* source_name,
		int line_no
            )
    {
	if ((source_name!=NULL)&&(strcmp(source_name, "asp")==0)) {
	    return co_translate_line(line_no);
	}
	else {
	    return line_no;
	}
    }
    %%
    ;
  };

  method  evaluate
  {
     doc
     %%
       This is one of the most important methods. It is responsible for evaluating 
       all java script code. 
       Method works like this:
       1. reads code for given script
	   if par_id==-2, no code is read from database
	   However, libraries are still used as described below.
       2. looks for //#uses: phrase
       3. reads all names in uses: phrase
       4. finds all libraries to include using select id from js_lib where name like from_uses_phrase
       5. takes all ids from point 4 and reads all trees for this ids
       6. creates one big string and passes it for execution
       
       REMARKS:
       there is no mechanism to detect repeated code in libs (will be in future)

	direct_code is used to pass code 'on fly' (not stored in database);

	prolog is executed before any libaries. it sould not depend on any
	    libraries.
	libraries are used in code from database and direct_code (and library
	    code is executed BEFORE database/direct code).

	CODE EXECUTION ORDER:
	prolog ("//#uses:..." in prolog is ignored)
	libraries (order???)
	direct_code OR code specified by par_id ("//#uses:..." describes libraries used)

	WARNING: If par_id is !=-2 (i.e. code is read from database 
	    AND direct_code is not NULL, then method behaviour is undefined 
	    and may change in the future versions.

	Library code fragments are evaluated separately; error messages are marked
	as follows:
	prolog      <=> error in prolog
	code        <=> error in submitted code (stored in database or direct) if NOT ASP
	asp         <=> error in submitted code if evaluated as ASP
	library <name>[<id>] <=> error in library named <name>, code fragment stored in js_code
	     with id=<id>
     %%
     ;
       
     id_permission __PERMISSION_EVALUATE;
     sec_type __METHOD_SEC_TYPE_OBJECT;

     __PARAMS_HEADER
     
     param prolog {type CO_STRING; optional 1; default 'prolog=ap_pstrdup(p,"");';};
     param asp {type CO_INTEGER; optional 1; default 'asp=0;';};
     param direct_code {type CO_STRING; optional 1; default 'direct_code=NULL;';};
     param direct_libs {type CO_STRING; optional 1; default 'direct_libs=NULL;';};
     param in_params {type CO_POINTER; optional 1; default 'in_params=NULL;';};
     param method_name {type CO_STRING; optional 1; default 'method_name=NULL;';};
     param target_name {type CO_STRING; optional 1; default 'target_name=NULL;';};
 
     declare language_c
     {
        C 'int nTuples;';
        C 'int id_js_code;';
        C 'int i,j;';
        C 'const char *expression_result;';
        C 'const char *function_name="";';
        C 'const char *function_arg="";';
        C 'const char *function_source;';
        C 'const char *param_name;';
        C 'const char *param_type;';
        C 'const char *param_default;';
        C 'int param_optional;';
        C 'const char *param_value;';
        C 'const char *ids_libs;';
        C 'const char *empty="";';
        C 'const char *id_js_lib;';
        C 'int id_js_lib_i;'; 
        C 'const char *uses;';
        C 'const char *uses_pat = "\n//#uses:";';
        C 'const char *uses_pat_begin = "//#uses:";';
        C 'const char *lib_name;';
        C 'int id_lib;';
        C 'int old_send_data;';
        C 'JSObject *globalObj;';
        C 'JSContext *jscx;';
        C 'JSBool jsok;';
        C 'JSScript* comp;';
        C 'jsval rval;';
        C 'const char * js_code;';
        C 'int id_catalog_type;';
        C 'sigjmp_buf sigenv_tmp;';
     };
     body language_c
     {
        C
        %%
    /* getting code for this js_code we use call to get method to open way for future changes 
       ie. we do not call select from database directly intentionaly */
	err=0;
    if (par_id != -2) {
	@co_getProp(par_area, '"js_code"',par_id,'"code"');
    }
    else {
	get_str="";
	par_id=0;
    }
    if(err==0)
    {
       js_code = get_str;
    }
    else
    {
       cou_sendMessage(cnx, ap_psprintf(p, "could not find code for id: %d",par_id));
       return -1;
    };      

    cou_sendMessage(cnx, "target:");
    cou_sendMessage(cnx, target_name);
    cou_sendMessage(cnx, "method:");
    cou_sendMessage(cnx, method_name);

    cox_initJSContext( cnx, false ); /* second parameter is discarded (false used) */
	/* 
	 * now, we have to take care of the context and make sure it is destroyed
	 * in any case, including errors (cou_error) in any function called, 
	 * in JavaScript, in native methods
	 */
    sigenv_tmp[0] = cnx->sigenv[0];
    if ( sigsetjmp(cnx->sigenv, CO_ERR_SAVE_SIGNALS) == 0) {

    jscx = cnx->jscnx;
    cjs_setCOInParams(jscx, in_params);
    cjs_setCOMethodName(jscx, method_name);
    cjs_setCOTargetName(jscx, target_name);

    globalObj = JS_GetGlobalObject(jscx);

    jsok = JS_EvaluateScript(jscx, globalObj, prolog, strlen(prolog),
                           "prolog", 1, &rval);
    if (!jsok) {
		cjs_dumpErrorMessages(jscx);
        cou_error( cnx, "js_code::evaluate: JS_ExecuteScript failed.");
    }

    /* now, if we got asp, it is time to switch line translation */
    if (asp!=0) {
		/* asp line translation from ap_parser.h */
		cjs_setCOLineTranslation(jscx, loc_lineTranslation );
    }

    if (direct_code) function_source = direct_code;
    else function_source = js_code;

    cou_sendMessage(cnx, "f source:");
    cou_sendMessage(cnx, function_source);

    /* set uses to library list */
    if (direct_libs) {
	uses = direct_libs;
    }
    else {
		/* is there //#uses: at the beginning of code? */
		if( (!strncmp(function_source, uses_pat_begin, strlen(uses_pat_begin))) ) {
	    	uses = function_source + strlen(uses_pat_begin);
			cou_sendMessage(cnx, "uses 1: ");
			cou_sendMessage(cnx, uses);
		} /* or inside at a beginning of a line? */
		else if( (uses=strstr(function_source, uses_pat)) ) {
	    	uses += strlen(uses_pat);
		}
    }

    if(uses)
    {
    //  function_beg_source = ap_pnstrdup(p,function_source, uses-function_source);
//      function_source = uses;
      uses = ap_getword_nulls_nc(r->pool, (char **)&uses, ';');
//      function_source += strlen(uses)+ strlen(uses_pat);
      @co_lcall(par_area,'"catalog_type"',getIdForClasses,par_area=>par_area,par_class=>'"catalog_type"',parent_class=>'"js_lib"',class_name=>par_class,par_id=>'"-1"');
      __GET_RETURNED_INTEGER(id_catalog_type)
      cou_sendMessage(cnx, "uses: ");
      cou_sendMessage(cnx, uses);
      i=0;           
      while( (lib_name = ap_getword(r->pool, &uses, ',')) )
      {
        cou_sendMessage(cnx, "lib_name: ");
        cou_sendMessage(cnx, lib_name);

        lib_name = cou_spaceTrim(cnx, lib_name);
	
	@co_filterByProp(par_area, '"js_lib"','"id"','"name"',lib_name);
	nTuples = cor_GetTupleCount(cnx, corec);
	for(i=0; i<nTuples; i++)
	{
	   cou_sendMessage(cnx, "poszczegolny lib_id: ");
       cou_sendMessage(cnx, cor_GetValue(cnx, corec,i,0));

  	   id_lib = atoi(cor_GetValue(cnx, corec,i,0));
	
       /* getting library code */
       if(id_lib>0)
       {
	     @co_getProp(par_area, '"js_lib"', id_lib,'"id_catalog_type"');
         if (id_catalog_type == atoi(get_str)) {
	        @co_getProp(par_area, '"js_lib"', id_lib,'"id_item"');
  	        id_js_lib_i = atoi(get_str);/* id_js_code */
	        if (id_js_lib_i>0) /* has code attached */
	        {
        	   	@co_getProp(par_area, par_class, id_js_lib_i, '"code"');
  	           	cou_logError(cnx, CO_DEBUG, "Evaluating: %s", get_str);
		       	jsok = JS_EvaluateScript(jscx, globalObj, get_str, strlen(get_str),
                           ap_psprintf(p,"library %s[%d]", lib_name, id_js_lib_i), 1, &rval);
		       	if (!jsok) {
    		       cou_error( cnx, "js_code::evaluate: JS_ExecuteScript failed.");
				}
	     	}
		 }
         @co_lcall(par_area,'"js_lib"',getTree,par_area=>par_area,par_class=>'"js_lib"',par_id=>id_lib,prop_name=>'"id_item"',deli=>'","',group_deli=>empty,callback=>empty);
         ids_libs = (char *) cou_getReturnValue(cnx, CO_STRING, &err);
         j=0;
         while( (id_js_lib = ap_getword(r->pool, &ids_libs, ',')) )
         {
               if(strlen(id_js_lib)>0)
               {
                  id_js_lib_i = atoi(id_js_lib);
                  cou_sendMessage(cnx, ap_psprintf(r->pool,"id: %d",id_js_lib_i));
                  @co_getProp(par_area, par_class, id_js_lib_i, '"code"');
//                  snprintf(JSOutBuf+strlen(JSOutBuf),JSBUFSIZE-strlen(JSOutBuf), "%s;", get_str);
		          jsok = JS_EvaluateScript(jscx, globalObj, get_str, strlen(get_str),
                           ap_psprintf(p,"library %s[%d]", lib_name, id_js_lib_i), 1, &rval);
		          if (!jsok) {
    		         cou_error( cnx, "js_code::evaluate: JS_ExecuteScript failed.");
				  }
               };
               if(strlen(ids_libs)==0) break;
               if(j++ >5) break; 
         }; // while id_js_lib
      }; // if id_lib>0
	}; // for   
    if(strlen(uses)==0) break;
    if(i++ >5) break;
    };//while lib_name
    }; // if uses
    

//    snprintf(JSOutBuf+strlen(JSOutBuf),JSBUFSIZE-strlen(JSOutBuf), "function __temporary_%d() {%s}", par_id,function_source);
    /* Wrap code so that variables are local */
    /* Do not change line numbers - unless you want to write 
       your own line translation function! */
	/* close function after new line - the last line of code could be // comment */
    snprintf(JSOutBuf,JSBUFSIZE, "function __temporary_%d() {%s\n}", par_id, function_source);
    cou_sendMessage(cnx, "source:");
    cou_sendMessage(cnx, JSOutBuf);

    if (asp==0) {
        jsok = JS_EvaluateScript(jscx, globalObj, JSOutBuf, strlen(JSOutBuf),
	               "code", 1, &rval);
    }
    else {
        /* we got asp in direct_code */
        jsok = JS_EvaluateScript(jscx, globalObj, JSOutBuf, strlen(JSOutBuf),
	               "asp", 1, &rval);
    }
    if (!jsok) {
        cou_error( cnx, "js_code::evaluate: JS_ExecuteScript failed.");
    }

    old_send_data = cnx->send_data;
    cnx->send_data = 0;
	/* JS takes care of cou_error calls from native methods */
    jsok = JS_CallFunctionName(jscx, globalObj, ap_psprintf(p,"__temporary_%d",par_id), 0, NULL, &rval);
    cnx->send_data = old_send_data;

    if (!jsok) {
        cou_error( cnx, "js_code::evaluate: JS_CallFunctionName failed.");
    }

    cou_setReturnValue( cnx,
                (void*)ap_pstrdup(p, JS_GetStringBytes(JS_ValueToString(jscx, rval))), CO_STRING);
    }
    else {
        /* error */
        /* restore error context */
        cnx->sigenv[0] = sigenv_tmp[0];
		cjs_dumpErrorMessages(jscx);
        JS_DestroyContext(jscx); /* destroy _current_ JavaScript context */
        cou_error( cnx, "js_code::evaluate: cou_error called.");
	}
	/* was OK */
    JS_DestroyContext(jscx); /* destroy _current_ context */
	/* OK, no context */
    expression_result = (char*)cou_getReturnValue( cnx, CO_STRING, &err );
    cou_logError( cnx, CO_LOG_DEBUG, "Result: %s", expression_result );
    cou_logError( cnx, CO_LOG_DEBUG, "Error: %d", err );                            
    /* restore error context */
    cnx->sigenv[0] = sigenv_tmp[0];

       %%
      ;
    };      
  };

  method  createScript
  {
     doc
     %%
       Create new script
       
       TODO:
	precompile script       
     %%
     ;
       
     id_permission __PERMISSION_CREATE;
     sec_type __METHOD_SEC_TYPE_CLASS;
 
     __PARAMS_HEADER
     
     param code {type CO_STRING; optional 1; default 'code=ap_pstrdup(p,"");';};
 
     declare language_c
     {
        C 'int id_js_code;';
        C 'char * strCodeEncoded;'; 
     };
     body language_c
     {
        C
        %%
//    
        @co_call(par_area, par_class, "create", par_area=>par_area, par_class=>'"js_code"', par_id=>-1);
        
	__GET_RETURNED_INTEGER(id_js_code)
	
// set permissions

	@co_call(par_area, par_class, "permissionOn", par_area=>par_area, par_class=>'"js_code"', par_id=>id_js_code, id_permission=>__PERMISSION_EVALUATE);
	@co_call(par_area, par_class, "permissionOn", par_area=>par_area, par_class=>'"js_code"', par_id=>id_js_code, id_permission=>__PERMISSION_GET);
	@co_call(par_area, par_class, "permissionOn", par_area=>par_area, par_class=>'"js_code"', par_id=>id_js_code, id_permission=>__PERMISSION_UPDATE);
	
	strCodeEncoded = cou_buildEncodedSQL(cnx, code);
	@co_call(par_area, par_class, "update", par_area=>par_area, par_class=>'"js_code"', par_id=>id_js_code, par_prop=>'"code"', new_value=>strCodeEncoded, version=>0);
	
    
	__RETURN_INTEGER( id_js_code)
    
	cou_logError( cnx, CO_LOG_DEBUG, "Result: %d", id_js_code );


          {
            const char* class_names[] = {"js_code",NULL};
            int object_ids[] = {id_js_code};

            cou_queueActionX( cnx, "js_code", class_names, object_ids,
                CO_OP_INSERT,
                "id", ap_psprintf(cnx->request->pool, "%d", id_js_code),
                "version", "1",
                "code", code, NULL );
          }

       %%
      ;
    };      
  };

};

//END OF CLASS CATALOG